from __future__ import absolute_import, print_function, unicode_literals

import logging

from zenaton.errors import ZenatonError
from zenaton.serialization import (
    get_class_path,
    get_properties,
    serialize,
)
from zenaton.task import Task
from zenaton.tasks.wait import Wait
from zenaton.workflow import Workflow


DEFAULT_TIMEOUT = (2 ** 31) - 1


log = logging.getLogger(__name__)


class OutputBox:

    def __init__(self, executable, position):
        self.executable = executable
        self.position_str = str(position)

    def __repr__(self):
        return 'OutputBox(executable={}, position={})'.format(
            self.executable,
            self.position_str,
        )

    def get_work_item(self):
        log.debug('OutputBox.get_work_item()')
        data = {
            'type': self.type,
            'name': self.name,
            'position': self.position_str,
            'input': self.input,
            'timeout': self.timeout,
        }
        if hasattr(self.executable, 'get_event'):
            data['event'] = self.executable.get_event()
        log.debug('work item = %s', data)
        return data

    @property
    def type(self):
        if self.is_task():
            return 'task'
        if self.is_workflow():
            return 'workflow'
        if self.is_wait():
            return 'wait'
        if self.is_wait_while():
            return 'while'
        raise ZenatonError('Unknown type')

    @property
    def name(self):
        if self.is_wait():
            return 'Wait'
        else:
            return get_class_path(self.executable.__class__)

    @property
    def input(self):
        if self.is_wait():
            return None
        else:
            return serialize(get_properties(self.executable))

    @property
    def timeout(self):
        if self.is_wait() or self.is_wait_while():
            return self.executable.get_timeout_timestamp()
        else:
            if hasattr(self.executable, 'get_timeout'):
                return self.executable.get_timeout()
            else:
                return DEFAULT_TIMEOUT

    def is_task(self):
        return isinstance(self.executable, Task)

    def is_workflow(self):
        return isinstance(self.executable, Workflow)

    def is_wait(self):
        return isinstance(self.executable, Wait)

    def is_wait_while(self):
        return False  # FIXME



from __future__ import absolute_import, print_function, unicode_literals

from zenaton.testing.compat import MagicMock, patch


@patch('sys.exit')
@patch('zenaton.scripts.init.init_microserver')
@patch('zenaton.scripts.init.setup_logging')
def test_main(mock_setup_logging, mock_init, mock_exit):
    from zenaton.scripts.init import main

    main([])

    mock_setup_logging.assert_called_once_with(debug=False, verbose=False)
    mock_init.assert_called_once_with(base_url=None)
    mock_exit.assert_called_once_with(0)


@patch('sys.exit')
@patch('zenaton.scripts.init.init_microserver')
@patch('zenaton.scripts.init.setup_logging')
def test_main_override_url(mock_setup_logging, mock_init, mock_exit):
    from zenaton.scripts.init import main

    main(['--microserver-url=http://localhost:4002/'])

    mock_setup_logging.assert_called_once_with(debug=False, verbose=False)
    mock_init.assert_called_once_with(base_url='http://localhost:4002/')
    mock_exit.assert_called_once_with(0)


@patch('sys.exit')
@patch('zenaton.scripts.init.init_microserver')
@patch('zenaton.scripts.init.setup_logging')
def test_main(mock_setup_logging, mock_init, mock_exit):
    from zenaton.scripts.init import main

    mock_init.side_effect = RuntimeError

    main([])

    mock_exit.assert_called_once_with(1)


@patch('zenaton.scripts.init.MicroserverAPI')
@patch('zenaton.scripts.init.Config')
def test_init_microserver(mock_config_class, mock_microserver_class):
    from zenaton.scripts.init import init_microserver

    mock_config_class.return_value.to_dict.return_value = {'foo': 'bar'}

    mock_microserver_instance = MagicMock()
    mock_microserver_class.return_value = mock_microserver_instance

    init_microserver(base_url='http://localhost:4002/')

    mock_microserver_class.assert_called_once_with(
        base_url='http://localhost:4002/',
    )
    mock_microserver_instance.send_config.assert_called_once_with(
        config={'foo': 'bar'}
    )



from __future__ import absolute_import, print_function, unicode_literals

import logging

from zenaton.errors import ZenatonError
from zenaton.output_box import OutputBox
from zenaton.position import Position
from zenaton.serialization import deserialize


log = logging.getLogger(__name__)


class DeciderWasModifiedError(ZenatonError):
    pass


class ScheduledBoxException(ZenatonError):
    pass


class WorkflowExecutor:

    def __init__(self, uuid, microserver):
        self.uuid = uuid
        self.microserver = microserver
        self.position = Position()

    def execute(self, workflow, executables, async_):
        log.debug(
            'WorkflowExecutor.execute(workflow=%s, executables=%s, async_=%s)',
            workflow,
            executables,
            async_,
        )

        boxes = self._build_boxes(executables, async_)
        resp = self.microserver.execute(self.uuid, boxes)

        log.debug('WorkflowExecutor.execute status=%s', resp['status'])

        if resp['status'] == 'modified':
            raise DeciderWasModifiedError(
                "Error: your workflow has changed - please use versioning"
            )

        if async_:
            return

        if resp['status'] == 'scheduled':
            raise ScheduledBoxException

        if resp['status'] == 'completed':
            workflow.update_properties(
                properties=deserialize(resp['properties'])
            )
            outputs = [deserialize(output) for output in resp['outputs']]
            log.debug('WorkflowExecutor.execute returning %s', outputs)
            return outputs if len(outputs) > 1 else outputs[0]

        raise ZenatonError(
            'InputBox with unknown status at position {}'.format(
                self.position,
            )
        )

    def _build_boxes(self, executables, async_):
        log.debug(
            'WorkflowExecutor._build_boxes(executables=%s, async_=%s)',
            executables,
            async_,
        )

        boxes = []

        for executable in executables:
            if async_:
                self.position.next_async()
            elif len(executables) > 1:
                self.position.next_parallel()
            else:
                self.position.next()
            box = OutputBox(executable, position=self.position)
            boxes.append(box)

        log.debug('boxes = %s)', boxes)
        return boxes



from __future__ import absolute_import, print_function, unicode_literals

from time import sleep

from client import client
from event.order_workflow import OrderWorkflow
from event.address_updated_event import AddressUpdatedEvent


def main():

    workflow = OrderWorkflow(
        item={
            'name': 'shirt',
        },
        address='1600 Pennsylvania Ave NW, Washington, DC 20500, USA',
    )

    # Direct synchronous execution
    # workflow.handle()

    # Execution through Zenaton
    instance = client.start(workflow)
    print('Launched workflow instance {}'.format(instance.id))

    sleep(2)

    instance.send_event(AddressUpdatedEvent(
        address='One Infinite Loop Cupertino, CA 95014',
    ))
    print('Event sent!')


if __name__ == '__main__':
    main()



from __future__ import absolute_import, print_function, unicode_literals

import os

from zenaton import Client


APP_ID = os.environ['ZENATON_APP_ID']
API_TOKEN = os.environ['ZENATON_API_TOKEN']
APP_ENV = os.environ['ZENATON_APP_ENV']
BASE_URL = os.getenv('ZENATON_BASE_URL')

client = Client(APP_ID, API_TOKEN, APP_ENV, BASE_URL)



from __future__ import absolute_import, print_function, unicode_literals

from zenaton import Workflow

from .book_car import BookCar
from .send_booking_confirmation import SendBookingConfirmation


class CarBookingWorkflow(Workflow):

    def __init__(self, request):
        self.request = request

    def handle(self):
        request = self.execute(BookCar(self.request))
        self.execute(SendBookingConfirmation(request))



from __future__ import absolute_import, print_function, unicode_literals

from zenaton import Workflow, Wait

from .send_welcome_email1 import SendWelcomeEmail1
from .send_welcome_email2 import SendWelcomeEmail2
from .send_welcome_email3 import SendWelcomeEmail3


class WelcomeWorkflow(Workflow):

    def __init__(self, user):
        self.user = user

    def handle(self):
        self.execute(SendWelcomeEmail1(self.user['email']))
        self.execute(Wait(seconds=4))
        self.execute(SendWelcomeEmail2(self.user['email']))
        self.execute(Wait(seconds=4))
        self.execute(SendWelcomeEmail3(self.user['email']))



from __future__ import absolute_import, print_function, unicode_literals

import json
import logging
from collections import OrderedDict

import requests

from zenaton.config import PROGRAMMING_LANGUAGE
from zenaton.errors import (
    APIError,
    UnauthenticatedError,
)
from zenaton.version import __version__


BASE_URL = 'https://zenaton.com/api'


log = logging.getLogger(__name__)


class API:
    """
    Zenaton.io API client.

    Arguments:
        app_id (str): The ID of the app you created on https://zenaton.com/.
            (Note that this is not the name you chose for your app
            (e.g. `test`), but the unique ID given by Zenaton, which
            looks like ``WMBAVMXSHQ``.
        token (str): An API access token that you created on
            https://zenaton.com/.
        environment (str): An arbitrary string that you can use to
            differentiate between environments, such as `dev`, `staging`
            or `production`.
        base_url (str, optional): Used to override the base URL of the
            Zenaton API server.
    """

    def __init__(self, app_id, token, environment, base_url=None):
        self.app_id = app_id
        self.token = token
        self.environment = environment
        self.base_url = base_url or BASE_URL
        self.session = requests.Session()

    def start_workflow(self, name, data, custom_id=None):
        return self._post_json(
            self.base_url + '/instances',
            data={
                'name': name,
                'data': data,
                'custom_id': custom_id,
                'programming_language': PROGRAMMING_LANGUAGE,
            },
        )

    def get_workflow_instance(self, id, name):
        # Feels like a workflow instance's ID should be unique
        # within an (app_id, app_env) context. Why do we need
        # to send the name too?
        return self._get_json(
            self.base_url + '/instances/' + id,
            params=OrderedDict([
                ('name', name),
                ('programming_language', PROGRAMMING_LANGUAGE),
            ])
        )

    def update_workflow_instance(self, id, name, state):
        return self._put_json(
            self.base_url + '/instances/' + id,
            data={
                'name': name,
                'programming_language': PROGRAMMING_LANGUAGE,
                'state': state,
            }
        )

    def send_event(self, custom_id, workflow_name, event_name, event_input):
        return self._post_json(
            self.base_url + '/events',
            data={
                'custom_id': custom_id,
                'event_input': event_input,
                'event_name': event_name,
                'name': workflow_name,
                'programming_language': PROGRAMMING_LANGUAGE,
            }
        )

    def _get_json(self, url, params=None):
        return self._request_json('GET', url, params=params)

    def _post_json(self, url, data, params=None):
        return self._request_json('POST', url, data=data, params=params)

    def _put_json(self, url, data, params=None):
        return self._request_json('PUT', url, data=data, params=params)

    def _request_json(self, method, url, data=None, params=None):
        if params is None:
            params = OrderedDict()
        params.update(OrderedDict([
            ('api_token', self.token),
            ('app_env', self.environment),
            ('app_id', self.app_id),
        ]))
        resp = self.session.request(
            method,
            url,
            params=params,
            headers={
                'Accept': 'application/json',
                'Content-Type': 'application/json',
                'User-Agent': 'zenaton-python/{}'.format(__version__),
            },
            data=json.dumps(data),
        )

        if resp.status_code == 401:
            raise UnauthenticatedError("Check your API token")

        log.debug("Request headers: %s", resp.request.headers)
        log.debug("Request body: %s", resp.request.body)
        log.debug("Response status: %s", resp.status_code)
        log.debug("Response headers: %s", resp.headers)
        log.debug("Response body: %s", resp.text)

        json_data = resp.json()

        if set(json_data.keys()) == {'error'}:
            raise APIError(json_data['error'])

        return json_data



from __future__ import absolute_import, print_function, unicode_literals

from .client import Client  # noqa
from .event import Event  # noqa
from .task import Task  # noqa
from .tasks.wait import Wait  # noqa
from .workflow import Workflow  # noqa



from __future__ import absolute_import, print_function, unicode_literals

import logging

from six import binary_type, text_type

from zenaton.api import API
from zenaton.errors import ZenatonError
from zenaton.proxy import WorkflowProxy
from zenaton.serialization import (
    get_class_path,
    get_properties,
    serialize,
)
from zenaton.utils import has_method


MAX_ID_BYTES = 191


log = logging.getLogger(__name__)


class Client:
    """
    Zenaton API client.

    Arguments:
        app_id (str): The ID of the app you created on https://zenaton.com/.
            (Note that this is not the name you chose for your app
            (e.g. `test`), but the unique ID given by Zenaton, which
            looks like ``WMBAVMXSHQ``.
        token (str): An API access token that you created on
            https://zenaton.com/.
        environment (str): An arbitrary string that you can use to
            differentiate between environments, such as `dev`, `staging`
            or `production`.
        base_url (str, optional): Used to override the base URL of the
            Zenaton API server.
    """

    def __init__(self, app_id, token, environment, base_url=None):
        self.api = API(app_id, token, environment, base_url)

    def start(self, workflow):
        """
        Start execution of a new workflow instance.
        """
        if not has_method(workflow, 'handle'):
            raise ZenatonError('Workflow must implement the handle() method')

        name = get_class_path(workflow.__class__)

        response_data = self.api.start_workflow(
            name=name,
            data=serialize(get_properties(workflow)),
            custom_id=self._get_custom_id(workflow),
        )

        log.debug("started workflow: %s", response_data)

        return WorkflowProxy(
            api=self.api,
            id=response_data['custom_id'],
            name=name,
        )

    def _get_custom_id(self, workflow):
        if hasattr(workflow, 'get_id') and callable(workflow.get_id):
            custom_id = workflow.get_id()
            if custom_id is None:
                return None
            if isinstance(custom_id, binary_type):
                custom_id = custom_id.decode()
            if not isinstance(custom_id, text_type):
                raise ZenatonError('Workflow ID must be a string')
            if len(custom_id.encode('utf-8')) > MAX_ID_BYTES:
                raise ZenatonError('Workflow ID is too long')
            log.info(custom_id)
            return custom_id
        return None



"""
Example from the README
"""
from __future__ import absolute_import, print_function, unicode_literals

import logging
import os

from zenaton import Client, Task, Workflow


log = logging.getLogger(__name__)


class Echo(Task):

    def __init__(self, value):
        self.value = value

    def handle(self):
        log.debug('Echo.handle()')
        print(self.value)


class MyWorkflow(Workflow):

    def __init__(self, count):
        self.count = count

    def handle(self):
        log.debug('MyWorkflow.handle() start')
        while self.count > 0:
            self.execute(Echo(self.count))
            self.count -= 1
        log.debug('MyWorkflow.handle() end')


def main():
    logging.basicConfig(level=logging.DEBUG)

    # Create a simple workflow
    workflow = MyWorkflow(count=1)

    # Ask Zenaton to start an instance of this workflow
    client = Client(
        app_id=os.environ['ZENATON_APP_ID'],
        token=os.environ['ZENATON_API_TOKEN'],
        environment=os.environ['ZENATON_APP_ENV'],
        base_url=os.getenv('ZENATON_BASE_URL'),
    )
    instance = client.start(workflow)
    print(instance.id)


if __name__ == '__main__':
    main()



from __future__ import absolute_import, print_function, unicode_literals

import json
import logging
import os
from collections import OrderedDict

from zenaton.compat import JSONDecodeError
from zenaton.errors import ZenatonError
from zenaton.instantiation import instantiate_class


log = logging.getLogger(__name__)


class SerializeError(ZenatonError):
    pass


class DeserializeError(ZenatonError):
    pass


class InvalidPropertyError(ZenatonError):
    pass


class InvalidClassError(ZenatonError):
    pass


class Serializable:
    """
    Mixin for (de)serializable classes.
    """

    def __init__(self, **properties):
        """
        Default constructor initializing properties from keyword args
        """
        self.update_properties(properties)

    def update_properties(self, properties):
        for name, value in properties.items():
            if name.startswith('_'):
                msg = "Invalid property name '{}'".format(name)
                raise InvalidPropertyError(msg)
            setattr(self, name, value)

    @classmethod
    def create(cls, class_path, properties):
        obj = instantiate_class(class_path, properties)
        if not isinstance(obj, cls):
            raise InvalidClassError('Not a {}'.format(cls.__name__))
        return obj


def serialize(obj):
    try:
        return json.dumps(obj, default=_default)
    except TypeError:
        msg = 'Cannot serialize {}'.format(obj)
        log.exception(msg)
        raise SerializeError(msg)


def _default(obj):
    if isinstance(obj, Serializable):
        return OrderedDict([
            ('name', get_class_path(obj.__class__)),
            ('properties', get_properties(obj)),
        ])
    raise TypeError


def get_properties(obj):
    return {
        key: value
        for key, value in vars(obj).items()
        if not key.startswith('_')
    }


def get_class_path(class_):
    return '{}:{}'.format(
        get_module_name(class_),
        class_.__name__,
    )


def get_module_name(class_):
    if class_.__module__ == '__main__':
        import __main__
        module_path = os.path.normpath(__main__.__file__)
        root, ext = os.path.splitext(module_path)
        return '.'.join(root.split(os.sep))
    return class_.__module__


def deserialize(text):
    """
    Deserialize an object from the API
    """
    if text is None:
        return None
    try:
        return json.loads(
            text,
            object_hook=_object_hook,
        )
    except JSONDecodeError as exc:
        raise DeserializeError("Invalid JSON: {}".format(exc))


def _object_hook(obj):
    if set(obj.keys()) == {'name', 'properties'}:
        class_path = obj['name']
        return instantiate_class(
            class_path=class_path,
            properties=obj['properties'],
        )
    return obj



"""
End-to-end test with fake remote and microserver APIs
"""
from __future__ import absolute_import, print_function, unicode_literals

import json
import logging

import responses

from zenaton.testing.compat import patch


log = logging.getLogger(__name__)


@patch('sys.exit')
def test_zenaton_slave_run_decider(mock_exit, capsys):
    from zenaton.scripts.slave import main

    with responses.RequestsMock() as rsps:

        # Get a job to do => run a decision
        rsps.add(
            responses.GET,
            'http://localhost:4001/jobs/AD489DB1',
            status=200,
            json={
                'action': 'DecisionScheduled',
                'uuid': '95BBD85D-CC3F-4401-9337-D329FC4CDBEF',
            }
        )

        # === 1st run of the decider: schedule a task

        # {"action": "start"}
        rsps.add(
            responses.POST,
            'http://localhost:4001/decisions/95BBD85D-CC3F-4401-9337-D329FC4CDBEF',
            json={
                'properties': '{"count": 1}',
                'name': 'zenaton.testing.examples.readme:MyWorkflow',
                'event': None,
            },
        )

        # {"action": "execute"}
        rsps.add(
            responses.POST,
            'http://localhost:4001/decisions/95BBD85D-CC3F-4401-9337-D329FC4CDBEF',
            json={
                'status': 'scheduled',
            },
        )

        # {"action": "terminate"}
        rsps.add(
            responses.POST,
            'http://localhost:4001/decisions/95BBD85D-CC3F-4401-9337-D329FC4CDBEF',
            json={
                'properties': '{"count": 0}',
                'status': 'running',
            },
        )

        # === 2nd run of the decider: task is completed

        # {"action": "start"}
        rsps.add(
            responses.POST,
            'http://localhost:4001/decisions/95BBD85D-CC3F-4401-9337-D329FC4CDBEF',
            json={
                'properties': '{"count": 1}',
                'name': 'zenaton.testing.examples.readme:MyWorkflow',
                'event': None,
            },
        )

        # {"action": "execute"}
        rsps.add(
            responses.POST,
            'http://localhost:4001/decisions/95BBD85D-CC3F-4401-9337-D329FC4CDBEF',
            json={
                'status': 'completed',
                'properties': '{"count": 1}',
                'outputs': ['null'],
            },
        )

        # {"action": "terminate"}
        rsps.add(
            responses.POST,
            'http://localhost:4001/decisions/95BBD85D-CC3F-4401-9337-D329FC4CDBEF',
            json={
                'properties': '{"count": 0}',
                'status': 'completed',
            },
        )

        # === 3rd run of the decider: we're done

        # {"action": "start"}
        rsps.add(
            responses.POST,
            'http://localhost:4001/decisions/95BBD85D-CC3F-4401-9337-D329FC4CDBEF',
            body='[]',
            content_type='application/json',
        )

        main(['AD489DB1', '8CF89842'])

        assert len(rsps.calls) == 8

        assert rsps.calls[0].request.url == 'http://localhost:4001/jobs/AD489DB1?slave_id=8CF89842'

        # == 1st run of the decider: schedule a task

        assert rsps.calls[1].request.url == 'http://localhost:4001/decisions/95BBD85D-CC3F-4401-9337-D329FC4CDBEF'
        assert json.loads(rsps.calls[1].request.body) == {
            'action': 'start',
        }

        assert rsps.calls[2].request.url == 'http://localhost:4001/decisions/95BBD85D-CC3F-4401-9337-D329FC4CDBEF'
        assert json.loads(rsps.calls[2].request.body) == {
            'action': 'execute',
            'works': [
                {
                    'input': '{"value": 1}',
                    'name': 'zenaton.testing.examples.readme:Echo',
                    'position': '1',
                    'timeout': 2147483647,
                    'type': 'task',
                }
            ],
        }

        assert rsps.calls[3].request.url == 'http://localhost:4001/decisions/95BBD85D-CC3F-4401-9337-D329FC4CDBEF'
        assert json.loads(rsps.calls[3].request.body) == {
            'action': 'terminate',
            'properties': '{"count": 1}',
            'status': 'running',
        }

        # === 2nd run of the decider: task is completed

        assert rsps.calls[4].request.url == 'http://localhost:4001/decisions/95BBD85D-CC3F-4401-9337-D329FC4CDBEF'
        assert json.loads(rsps.calls[4].request.body) == {
            'action': 'start',
        }

        assert rsps.calls[5].request.url == 'http://localhost:4001/decisions/95BBD85D-CC3F-4401-9337-D329FC4CDBEF'
        assert json.loads(rsps.calls[5].request.body) == {
            'action': 'execute',
            'works': [
                {
                    'input': '{"value": 1}',
                    'name': 'zenaton.testing.examples.readme:Echo',
                    'position': '1',
                    'timeout': 2147483647,
                    'type': 'task',
                }
            ],
        }

        assert rsps.calls[6].request.url == 'http://localhost:4001/decisions/95BBD85D-CC3F-4401-9337-D329FC4CDBEF'
        assert json.loads(rsps.calls[6].request.body) == {
            'action': 'terminate',
            'properties': '{"count": 0}',
            'output': 'null',
            'status': 'completed',
        }

        # === 3rd run of the decider: we're done

        assert rsps.calls[7].request.url == 'http://localhost:4001/decisions/95BBD85D-CC3F-4401-9337-D329FC4CDBEF'
        assert json.loads(rsps.calls[7].request.body) == {
            'action': 'start',
        }


    # out, err = capsys.readouterr()
    # assert out == '10\n'
    # assert err == ''

    mock_exit.assert_called_once_with(0)



from __future__ import absolute_import, print_function, unicode_literals

from textwrap import dedent
import sys

import pytest

from zenaton.testing.compat import MagicMock, patch


@pytest.fixture
def microserver():
    from zenaton.microserver import MicroserverAPI
    microserver = MicroserverAPI()
    microserver.session = MagicMock()
    return microserver


class TestMicroserverBaseURL:

    def test_default_base_url(self):
        from zenaton.microserver import MicroserverAPI
        microserver = MicroserverAPI()
        assert microserver.base_url == 'http://localhost:4001'

    def test_explicit_none_base_url(self):
        from zenaton.microserver import MicroserverAPI
        microserver = MicroserverAPI(base_url=None)
        assert microserver.base_url == 'http://localhost:4001'

    def test_override_base_url(self):
        from zenaton.microserver import MicroserverAPI
        microserver = MicroserverAPI(base_url='http://localhost:4002')
        assert microserver.base_url == 'http://localhost:4002'


class TestMicroServerSendConfig:

    def test_send_config(self, microserver):
        microserver._post_json = MagicMock()
        microserver._post_json.return_value = {
            'foo': 'bar',
        }

        res = microserver.send_config({'a': 1})

        microserver._post_json.assert_called_once_with(
            url='http://localhost:4001/configuration',
            data={'a': 1},
        )
        assert res == {
            'foo': 'bar',
        }


class TestMicroServerAskForJob:

    def test_ask_for_job(self, microserver):
        microserver._get_json = MagicMock()
        microserver._get_json.return_value = {
            'action': 'DecisionScheduled',
        }

        res = microserver.ask_for_job(
            instance_id='A21CA381-6F97-400B-B10C-3AAD198F33A9',
            slave_id='C5EDBCD0-DEBE-4C3C-816B-883DC4306B7B',
        )

        microserver._get_json.assert_called_once_with(
            url='http://localhost:4001/jobs/A21CA381-6F97-400B-B10C-3AAD198F33A9',
            params={
                'slave_id': 'C5EDBCD0-DEBE-4C3C-816B-883DC4306B7B',
            }
        )
        assert res == {
            'action': 'DecisionScheduled',
        }


class TestMicroserverDecisionStatus:

    def test_get_workflow_to_execute(self, microserver):
        microserver._post_decision_status = MagicMock()
        microserver._post_decision_status.return_value = {
            'foo': 'bar',
        }

        res = microserver.get_workflow_to_execute(
            '7997851A-E0D6-4E45-A511-AE0159D4FDAB',
        )

        microserver._post_decision_status.assert_called_once_with(
            '7997851A-E0D6-4E45-A511-AE0159D4FDAB',
            {
                'action': 'start',
            }
        )
        assert res == {
            'foo': 'bar',
        }

    def test_execute(self, microserver):
        microserver._post_decision_status = MagicMock()
        microserver._post_decision_status.return_value = {
            'foo': 'bar',
        }

        res = microserver.execute(
            uuid='7997851A-E0D6-4E45-A511-AE0159D4FDAB',
            boxes=[]
        )

        microserver._post_decision_status.assert_called_once_with(
            '7997851A-E0D6-4E45-A511-AE0159D4FDAB',
            {
                'action': 'execute',
                'works': []
            }
        )
        assert res == {
            'foo': 'bar',
        }

    def test_notify_decision_complete(self, microserver):
        microserver._post_decision_status = MagicMock()
        microserver._post_decision_status.return_value = {
            'foo': 'bar',
        }

        res = microserver.notify_decision_complete(
            '7997851A-E0D6-4E45-A511-AE0159D4FDAB',
            {'a': 1},
        )

        microserver._post_decision_status.assert_called_once_with(
            '7997851A-E0D6-4E45-A511-AE0159D4FDAB',
            {
                'action': 'terminate',
                'status': 'running',
                'properties': '{"a": 1}',
            }
        )
        assert res == {
            'foo': 'bar',
        }

    def test_notify_decision_branch_complete(self, microserver):
        microserver._post_decision_status = MagicMock()
        microserver._post_decision_status.return_value = {
            'foo': 'bar',
        }

        res = microserver.notify_decision_branch_complete(
            '7997851A-E0D6-4E45-A511-AE0159D4FDAB',
            {'a': 1},
            42,
        )

        microserver._post_decision_status.assert_called_once_with(
            '7997851A-E0D6-4E45-A511-AE0159D4FDAB',
            {
                'action': 'terminate',
                'status': 'completed',
                'properties': '{"a": 1}',
                'output': '42',
            }
        )
        assert res == {
            'foo': 'bar',
        }

    def test_notify_decision_user_error(self, microserver):
        try:
            raise RuntimeError('foo')
        except:
            etype, value, tb = sys.exc_info()

        microserver._notify_decision_error = MagicMock()
        microserver._notify_decision_error.return_value = {
            'foo': 'bar',
        }

        res = microserver.notify_decision_user_error(
            '7997851A-E0D6-4E45-A511-AE0159D4FDAB',
            etype,
            value,
            tb,
        )

        microserver._notify_decision_error.assert_called_once_with(
            'failed',
            '7997851A-E0D6-4E45-A511-AE0159D4FDAB',
            etype,
            value,
            tb,
        )
        assert res == {
            'foo': 'bar',
        }

    def test_notify_decision_internal_error(self, microserver):
        from zenaton.errors import ZenatonError

        try:
            raise ZenatonError('foo')
        except ZenatonError:
            etype, value, tb = sys.exc_info()

        microserver._notify_decision_error = MagicMock()
        microserver._notify_decision_error.return_value = {
            'foo': 'bar',
        }

        res = microserver.notify_decision_internal_error(
            '7997851A-E0D6-4E45-A511-AE0159D4FDAB',
            etype,
            value,
            tb,
        )

        microserver._notify_decision_error.assert_called_once_with(
            'zenatonFailed',
            '7997851A-E0D6-4E45-A511-AE0159D4FDAB',
            etype,
            value,
            tb,
        )
        assert res == {
            'foo': 'bar',
        }

    def test_notify_decision_error(self, microserver):
        try:
            raise RuntimeError('foo')
        except:
            etype, value, tb = sys.exc_info()

        dummy_stacktrace = dedent('''\
            Traceback (most recent call last):
              File "/path/to/module.py", line 42, in some_function
                raise RuntimeError(\'foo\')
            zenaton.errors.RuntimeError: foo
        ''')

        microserver._post_decision_status = MagicMock()
        microserver._post_decision_status.return_value = {
            'foo': 'bar',
        }

        with patch('zenaton.microserver.format_exception') as mock_fmt_exc:
            mock_fmt_exc.return_value = dummy_stacktrace

            with patch('zenaton.microserver.utc_timestamp') as mock_timestamp:
                mock_timestamp.return_value = 1499358596

                res = microserver._notify_decision_error(
                    'reason',
                    '7997851A-E0D6-4E45-A511-AE0159D4FDAB',
                    etype,
                    value,
                    tb,
                )

        microserver._post_decision_status.assert_called_once_with(
            '7997851A-E0D6-4E45-A511-AE0159D4FDAB',
            {
                'action': 'terminate',
                'status': 'reason',
                'error_code': None,
                'error_name': 'RuntimeError',
                'error_message': 'foo',
                'error_stacktrace': dummy_stacktrace,
                'failed_at': 1499358596,
            }
        )
        assert res == {
            'foo': 'bar',
        }

    def test_post_decision_status(self, microserver):
        microserver._post_json = MagicMock()
        microserver._post_json.return_value = {
            'foo': 'bar',
        }

        res = microserver._post_decision_status('UUID', {'a': 1})

        microserver._post_json.assert_called_once_with(
            url='http://localhost:4001/decisions/UUID',
            data={'a': 1},
        )
        assert res == {
            'foo': 'bar',
        }


class TestMicroserverTaskStatus:

    def test_notify_task_success(self, microserver):
        microserver._post_task_status = MagicMock()
        microserver._post_task_status.return_value = {
            'foo': 'bar',
        }

        res = microserver.notify_task_success(
            uuid='197BAA87-9AF3-4B91-AAF9-A2DC7D65CB2E',
            hash_='1a08b1b6edbe8009ecf93969079efb92',
            result=42,
        )

        microserver._post_task_status.assert_called_once_with(
            '197BAA87-9AF3-4B91-AAF9-A2DC7D65CB2E',
            {
                'action': 'terminate',
                'status': 'completed',
                'output': '42',
                'duration': 0,
                'hash': '1a08b1b6edbe8009ecf93969079efb92',
            }
        )
        assert res == {
            'foo': 'bar',
        }

    def test_notify_task_user_error(self, microserver):
        try:
            raise RuntimeError('foo')
        except:
            etype, value, tb = sys.exc_info()

        microserver._notify_task_error = MagicMock()
        microserver._notify_task_error.return_value = {
            'foo': 'bar',
        }

        res = microserver.notify_task_user_error(
            '197BAA87-9AF3-4B91-AAF9-A2DC7D65CB2E',
            '1a08b1b6edbe8009ecf93969079efb92',
            etype,
            value,
            tb,
        )

        microserver._notify_task_error.assert_called_once_with(
            reason='failed',
            uuid='197BAA87-9AF3-4B91-AAF9-A2DC7D65CB2E',
            hash_='1a08b1b6edbe8009ecf93969079efb92',
            etype=etype,
            value=value,
            tb=tb,
        )
        assert res == {
            'foo': 'bar',
        }

    def test_notify_task_internal_error(self, microserver):
        from zenaton.errors import ZenatonError

        try:
            raise ZenatonError('foo')
        except ZenatonError:
            etype, value, tb = sys.exc_info()

        microserver._notify_task_error = MagicMock()
        microserver._notify_task_error.return_value = {
            'foo': 'bar',
        }

        res = microserver.notify_task_internal_error(
            '197BAA87-9AF3-4B91-AAF9-A2DC7D65CB2E',
            '1a08b1b6edbe8009ecf93969079efb92',
            etype,
            value,
            tb,
        )

        microserver._notify_task_error.assert_called_once_with(
            reason='zenatonFailed',
            uuid='197BAA87-9AF3-4B91-AAF9-A2DC7D65CB2E',
            hash_='1a08b1b6edbe8009ecf93969079efb92',
            etype=etype,
            value=value,
            tb=tb,
        )
        assert res == {
            'foo': 'bar',
        }

    def test_notify_task_error(self, microserver):
        try:
            raise RuntimeError('foo')
        except:
            etype, value, tb = sys.exc_info()

        dummy_stacktrace = dedent('''\
            Traceback (most recent call last):
              File "/path/to/module.py", line 42, in some_function
                raise RuntimeError(\'foo\')
            zenaton.errors.RuntimeError: foo
        ''')

        microserver._post_task_status = MagicMock()
        microserver._post_task_status.return_value = {
            'foo': 'bar',
        }

        with patch('zenaton.microserver.format_exception') as mock_fmt_exc:
            mock_fmt_exc.return_value = dummy_stacktrace

            with patch('zenaton.microserver.utc_timestamp') as mock_timestamp:
                mock_timestamp.return_value = 1499358596

                res = microserver._notify_task_error(
                    'reason',
                    '197BAA87-9AF3-4B91-AAF9-A2DC7D65CB2E',
                    '1a08b1b6edbe8009ecf93969079efb92',
                    etype,
                    value,
                    tb,
                )

        microserver._post_task_status.assert_called_once_with(
            '197BAA87-9AF3-4B91-AAF9-A2DC7D65CB2E',
            {
                'action': 'terminate',
                'status': 'reason',
                'error_code': None,
                'error_name': 'RuntimeError',
                'error_message': 'foo',
                'error_stacktrace': dummy_stacktrace,
                'failed_at': 1499358596,
                'hash': '1a08b1b6edbe8009ecf93969079efb92',
            }
        )
        assert res == {
            'foo': 'bar',
        }

    def test_post_task_status(self, microserver):
        microserver._post_json = MagicMock()
        microserver._post_json.return_value = {
            'foo': 'bar',
        }

        res = microserver._post_task_status('UUID', {'a': 1})

        microserver._post_json.assert_called_once_with(
            url='http://localhost:4001/works/UUID',
            data={'a': 1},
        )
        assert res == {
            'foo': 'bar',
        }


class TestMicroserverHTTP:

    def test_get_json(self, microserver):
        microserver._request_json = MagicMock()
        microserver._request_json.return_value = {'foo': 'bar'}

        microserver._get_json(
            url='http://localhost:4001/foo',
            params={'a': 1}
        )

        microserver._request_json.assert_called_once_with(
            'GET',
            'http://localhost:4001/foo',
            params={'a': 1},
        )

    def test_post_json(self, microserver):
        microserver._request_json = MagicMock()
        microserver._request_json.return_value = {'foo': 'bar'}

        microserver._post_json(
            url='http://localhost:4001/foo',
            data={'a': 1},
        )

        microserver._request_json.assert_called_once_with(
            'POST',
            'http://localhost:4001/foo',
            data={"a": 1},
            params=None,
        )

    def test_request_json(self, microserver):
        from zenaton.version import __version__

        microserver._request_json(
            'POST',
            'https://localhost:8080/api',
            data={'data': 1},
            params={'param': 1},
        )
        microserver.session.request.assert_called_once_with(
            'POST',
            'https://localhost:8080/api',
            params={
                'param': 1,
            },
            headers={
                'Accept': 'application/json',
                'Content-Type': 'application/json',
                'User-Agent': 'zenaton-python/{}'.format(__version__),
            },
            data='{"data": 1}',
        )
        resp = microserver.session.request.return_value
        resp.json.assert_called_once_with()



"""
End-to-end test with fake remote and microserver APIs
"""
from __future__ import absolute_import, print_function, unicode_literals

import responses


@responses.activate
def test_readme_example(monkeypatch):
    from zenaton.testing.examples.readme import main

    monkeypatch.setenv('ZENATON_APP_ID', 'APP_ID')
    monkeypatch.setenv('ZENATON_APP_ENV', 'APP_ENV')
    monkeypatch.setenv('ZENATON_API_TOKEN', 'API_TOKEN')
    monkeypatch.delenv('ZENATON_BASE_URL', raising=False)

    responses.add(
        responses.POST,
        'https://zenaton.com/api/instances',
        status=200,
        json={
            'msg': 'workflow zenaton.testing.examples.readme:MyWorkflow launched',
            'custom_id':'4zwjxjp0axwk4os0',
            'instance_id':'4zwjxjp0axwk4os0-pc132y8aflcs88g',
        },
    )

    main()

    assert len(responses.calls) == 1
    assert responses.calls[0].request.url == 'https://zenaton.com/api/instances?api_token=API_TOKEN&app_env=APP_ENV&app_id=APP_ID'



from __future__ import absolute_import, print_function, unicode_literals

import os

import pytest


@pytest.fixture
def config():
    from zenaton.config import Config

    return Config()


@pytest.mark.parametrize('env_var,attr_name', [
    ('ZENATON_APP_ID', 'app_id'),
    ('ZENATON_APP_ENV', 'app_env'),
    ('ZENATON_API_TOKEN', 'api_token'),
])
def test_required_environment_variable(config, env_var, attr_name, monkeypatch):

    monkeypatch.delenv(env_var, raising=False)
    with pytest.raises(KeyError):
        getattr(config, attr_name)


def test_absolute_path_to_worker_script(config):
    assert os.path.isabs(config.worker_script)



from __future__ import absolute_import, print_function, unicode_literals

from client import client
from wait.welcome_workflow import WelcomeWorkflow


def main():
    user = {
        'email': 'user@example.com',
    }
    workflow = WelcomeWorkflow(user)

    # Direct synchronous execution
    # workflow.handle()

    # Execution through Zenaton
    instance = client.start(workflow)
    print('Launched workflow instance {}'.format(instance.id))


if __name__ == '__main__':
    main()



from __future__ import absolute_import, print_function, unicode_literals

import pytest

from zenaton.task import Task


class MyTask(Task):

    def __init__(self, x):
        self.x = x

    def handle(self):
        return self.x + 1


@pytest.fixture
def workflow():
    from zenaton.workflow import Workflow

    return Workflow()


class TestExecuteSyncNormal:

    def test_execute_nothing(self, workflow):
        res = workflow.execute()

        assert res is None

    def test_execute_a_task(self, workflow):
        from zenaton.task import Task

        task = MyTask(x=1)

        res = workflow.execute(task)

        assert res == 2

    def test_execute_two_tasks(self, workflow):
        task1 = MyTask(x=1)
        task2 = MyTask(x=2)

        res = workflow.execute(task1, task2)

        assert res == [2, 3]

    def test_execute_not_a_task(self, workflow):
        from zenaton.workflow import NotExecutableError
        with pytest.raises(NotExecutableError):
            workflow.execute(42)


class TestExecuteAsyncNormal:

    def test_execute_nothing(self, workflow):
        res = workflow.execute_async()

        assert res is None

    def test_execute_a_task(self, workflow):
        from zenaton.task import Task

        task = MyTask(x=1)

        res = workflow.execute_async(task)

        assert res is None

    def test_execute_two_tasks(self, workflow):
        task1 = MyTask(x=1)
        task2 = MyTask(x=2)

        res = workflow.execute_async(task1, task2)

        assert res is None



from __future__ import absolute_import, print_function, unicode_literals

import logging


def setup_logging(verbose=False, debug=False):
    log_level = logging.WARN
    if verbose:
        log_level = logging.INFO
    if debug:
        log_level = logging.DEBUG
    logging.basicConfig(level=log_level)
    logging.getLogger('urllib3').setLevel(logging.WARN)



"""Zenaton worker slave.

Usage:
  zenaton_slave [options] [--microserver-url=URL] INSTANCE_ID SLAVE_ID
  zenaton_slave -h | --help

Options:
  -h --help     Show this screen.
  -v --verbose  Increase log level.
  -d --debug    Increase log level even more.
"""
from __future__ import absolute_import, print_function, unicode_literals

import json
import logging
import sys

from docopt import docopt

from zenaton.compat import JSONDecodeError
from zenaton.decider import Decider
from zenaton.errors import ZenatonError
from zenaton.microserver import MicroserverAPI
from zenaton.scripts.base import setup_logging
from zenaton.task import Task


log = logging.getLogger(__name__)


class InvalidPropertiesError(ZenatonError):
    pass


def main(argv=None):
    args = docopt(__doc__, argv=argv)
    setup_logging(verbose=args['--verbose'], debug=args['--debug'])
    try:
        process_single_job(
            instance_id=args['INSTANCE_ID'],
            slave_id=args['SLAVE_ID'],
            base_url=args['--microserver-url']
        )
        sys.exit(0)
    except Exception:
        logging.exception("Unexpected error")
        sys.exit(1)


def process_single_job(instance_id, slave_id, base_url=None):
    """
    Ask the microserver for a single job and run it
    """
    microserver = MicroserverAPI(base_url=base_url)
    resp = microserver.ask_for_job(instance_id=instance_id, slave_id=slave_id)
    if 'action' in resp:
        if resp['action'] == 'DecisionScheduled':
            run_decision(
                microserver=microserver,
                uuid=resp['uuid'],
            )
        elif resp['action'] == 'TaskScheduled':
            run_task(
                microserver=microserver,
                uuid=resp['uuid'],
                hash_=resp['hash'],
                class_path=resp['name'],
                properties=resp['input'],
            )
        else:
            raise NotImplementedError


def run_decision(microserver, uuid):
    log.debug('zenaton_slave:run_decision(uuid=%s)', uuid)

    decider = Decider(uuid=uuid, microserver=microserver)
    decider.launch()


def run_task(microserver, uuid, hash_, class_path, properties):
    log.debug(
        'zenaton_slave:run_task('
        'uuid=%s, hash_=%s, class_path=%s, properties=%s)',
        uuid, hash_, class_path, properties
    )

    try:
        properties = json.loads(properties)
    except JSONDecodeError:
        msg = "Invalid JSON value for properties: '{}'".format(properties)
        raise InvalidPropertiesError(msg)

    task = Task.create(class_path, properties)

    try:
        result = task.handle()
        microserver.notify_task_success(uuid, hash_, result)
    except ZenatonError:
        etype, value, tb = sys.exc_info()
        microserver.notify_task_internal_error(uuid, hash_, etype, value, tb)
        raise
    except Exception:
        etype, value, tb = sys.exc_info()
        microserver.notify_task_user_error(uuid, hash_, etype, value, tb)
        raise



from __future__ import absolute_import, print_function, unicode_literals

from zenaton import Workflow

from .send_invite import SendInvite


class SendInvitesWorkflow(Workflow):

    def __init__(self, invites):
        self.invites = invites

    def handle(self):
        for invite in self.invites:
            self.execute_async(SendInvite(invite))



from __future__ import absolute_import, print_function, unicode_literals

import json
import logging
from pprint import pformat
from traceback import format_exception

import requests

from zenaton.serialization import serialize
from zenaton.utils import utc_timestamp
from zenaton.version import __version__


BASE_URL = 'http://localhost:4001'


log = logging.getLogger(__name__)


class MicroserverAPI:

    def __init__(self, base_url=None):
        self.base_url = base_url or BASE_URL
        self.session = requests.Session()

    # ===== Initialization =====

    def send_config(self, config):
        return self._post_json(
            url=self.base_url + '/configuration',
            data=config,
        )

    # ===== Jobs =====

    def ask_for_job(self, instance_id, slave_id):
        return self._get_json(
            url=self.base_url + '/jobs/{}'.format(instance_id),
            params={
                'slave_id': slave_id,
            }
        )

    # ===== Decision status =====

    def get_workflow_to_execute(self, uuid):
        log.debug('MicroserverAPI.get_workflow_to_execute(uuid=%s)', uuid)
        return self._post_decision_status(
            uuid,
            {
                'action': 'start',
            }
        )

    def execute(self, uuid, boxes):
        log.debug('MicroserverAPI.execute(uuid=%s, boxes=%s)', uuid, boxes)
        work_items = []
        for box in boxes:
            work_items.append(box.get_work_item())
        return self._post_decision_status(
            uuid,
            {
                'action': 'execute',
                'works': work_items,
            }
        )

    def notify_decision_complete(self, uuid, properties):
        return self._post_decision_status(
            uuid,
            {
                'action': 'terminate',
                'status': 'running',
                'properties': serialize(properties),
            }
        )

    def notify_decision_branch_complete(self, uuid, properties, output):
        return self._post_decision_status(
            uuid,
            {
                'action': 'terminate',
                'status': 'completed',
                'properties': serialize(properties),
                'output': serialize(output),
            }
        )

    def notify_decision_user_error(self, uuid, etype, value, tb):
        return self._notify_decision_error(
            'failed', uuid, etype, value, tb
        )

    def notify_decision_internal_error(self, uuid, etype, value, tb):
        return self._notify_decision_error(
            'zenatonFailed', uuid, etype, value, tb
        )

    def _notify_decision_error(self, reason, uuid, etype, value, tb):
        stacktrace = ''.join(format_exception(etype, value, tb))
        return self._post_decision_status(
            uuid,
            {
                'action': 'terminate',
                'status': reason,
                'error_code': None,
                'error_name': etype.__name__,
                'error_message': str(value),
                'error_stacktrace': stacktrace,
                'failed_at': utc_timestamp(),
            }
        )

    def _post_decision_status(self, uuid, data):
        return self._post_json(
            url=self.base_url + '/decisions/' + str(uuid),
            data=data,
        )

    # ===== Task status =====

    def notify_task_success(self, uuid, hash_, result):
        log.debug(
            'notify_task_success(uuid=%s, hash_=%s, result=%s',
            uuid, hash_, repr(result),
        )
        return self._post_task_status(
            uuid,
            {
                'action': 'terminate',
                'status': 'completed',
                'output': serialize(result),
                'duration': 0,
                'hash': hash_,
            }
        )

    def notify_task_user_error(self, uuid, hash_, etype, value, tb):
        return self._notify_task_error(
            reason='failed',
            uuid=uuid,
            hash_=hash_,
            etype=etype,
            value=value,
            tb=tb,
        )

    def notify_task_internal_error(self, uuid, hash_, etype, value, tb):
        return self._notify_task_error(
            reason='zenatonFailed',
            uuid=uuid,
            hash_=hash_,
            etype=etype,
            value=value,
            tb=tb,
        )

    def _notify_task_error(self, reason, uuid, hash_, etype, value, tb):
        stacktrace = ''.join(format_exception(etype, value, tb))
        return self._post_task_status(
            uuid,
            {
                'action': 'terminate',
                'status': reason,
                'error_code': None,
                'error_name': etype.__name__,
                'error_message': str(value),
                'error_stacktrace': stacktrace,
                'failed_at': utc_timestamp(),
                'hash': hash_,
            }
        )

    def _post_task_status(self, uuid, data):
        return self._post_json(
            url=self.base_url + '/works/' + str(uuid),
            data=data,
        )

    # ==== HTTP communication =====

    def _get_json(self, url, params=None):
        log.debug("GET %s", url)
        return self._request_json('GET', url, params=params)

    def _post_json(self, url, data, params=None):
        log.debug("POST %s", url)
        return self._request_json('POST', url, data=data, params=params)

    def _request_json(self, method, url, data=None, params=None):
        resp = self.session.request(
            method,
            url,
            params=params,
            headers={
                'Accept': 'application/json',
                'Content-Type': 'application/json',
                'User-Agent': 'zenaton-python/{}'.format(__version__),
            },
            data=json.dumps(data),
        )

        log.debug("Request headers: %s", resp.request.headers)
        if resp.request.method != 'GET':
            log.debug("Request body: %s", format_json(resp.request.body))
        log.debug("Response status: %s", resp.status_code)
        log.debug("Response headers: %s", resp.headers)
        log.debug("Response body: %s", format_json(resp.text))

        return resp.json()


def format_json(data):
    try:
        return pformat(json.loads(data), indent=4)
    except:
        return data



from __future__ import absolute_import, print_function, unicode_literals

from collections import OrderedDict
import json

import pytest

from zenaton.event import Event
from zenaton.testing.compat import MagicMock


@pytest.fixture
def api():
    return MagicMock()


@pytest.fixture
def workflow(api):
    from zenaton.proxy import WorkflowProxy
    return WorkflowProxy(api, 'ID', 'NAME')


class DummyEvent(Event):

    def __init__(self, a):
        self.a = a


def test_workflow_proxy_send_event(api, workflow):
    event = DummyEvent(42)

    workflow.send_event(event)

    api.send_event.assert_called_once_with(
        custom_id='ID',
        workflow_name='NAME',
        event_name='test_proxy:DummyEvent',
        event_input=json.dumps(OrderedDict([
            ("name", "test_proxy:DummyEvent"),
            ("properties", {"a": 42}),
        ])),
    )


class TestProxySignals:

    def test_workflow_proxy_kill(self, workflow):
        from zenaton.proxy import Signal
        workflow._send_signal = MagicMock()
        workflow.kill()
        workflow._send_signal.assert_called_once_with(Signal.KILL)

    def test_workflow_proxy_pause(self, workflow):
        from zenaton.proxy import Signal
        workflow._send_signal = MagicMock()
        workflow.pause()
        workflow._send_signal.assert_called_once_with(Signal.PAUSE)

    def test_workflow_proxy_resume(self, workflow):
        from zenaton.proxy import Signal
        workflow._send_signal = MagicMock()
        workflow.resume()
        workflow._send_signal.assert_called_once_with(Signal.RUN)

    def test_workflow_proxy_send_signal(self, api, workflow):
        workflow._send_signal('SIGNAL')
        api.update_workflow_instance.assert_called_once_with(
            'ID',
            'NAME',
            'SIGNAL',
        )


# def test_workflow_proxy_get_properties(api, workflow):
#     api.get_workflow_instance.return_value = {
#         'instance': {
#             'a': 1,
#             'b': 2,
#         }
#     }

#     properties = workflow.get_properties()

#     api.get_workflow_instance.assert_called_once_with('ID', 'NAME')
#     assert properties == {
#         'a': 1,
#         'b': 2,
#     }



from __future__ import absolute_import, print_function, unicode_literals

from client import client
from asynchronous.send_invites_workflow import SendInvitesWorkflow


def main():

    notifications = ['Gilles', 'Julien', 'Oussama', 'Alice', 'Charlotte', 'Balthazar', 'Annabelle', 'Louis']
    workflow = SendInvitesWorkflow(notifications)

    # Direct synchronous execution
    # workflow.handle()

    # Execution through Zenaton
    instance = client.start(workflow)
    print('Launched workflow instance {}'.format(instance.id))


if __name__ == '__main__':
    main()



from __future__ import absolute_import, print_function, unicode_literals

from random import randint
from time import sleep

from zenaton import Task


class SendOrder(Task):

    def __init__(self, item, address):
        self.item = item
        self.address = address

    def handle(self):
        print('Sending {} to {}... '.format(self.item, self.address), end='')
        sleep(randint(5, 10))
        print('sent.')



from __future__ import absolute_import, print_function, unicode_literals

from collections import namedtuple
from contextlib import contextmanager
import logging
import sys

from zenaton.errors import ZenatonError
from zenaton.serialization import (
    deserialize,
    get_properties,
)
from zenaton.workflow import Workflow
from zenaton.workflow_executor import (
    ScheduledBoxException,
    WorkflowExecutor,
)


log = logging.getLogger(__name__)


Branch = namedtuple('Branch', ['name', 'properties', 'event'])


class Decider:

    def __init__(self, uuid, microserver):
        self.uuid = uuid
        self.microserver = microserver

    def launch(self):
        log.debug('Decider.launch()')
        while True:
            branch = self.get_next_workflow_branch()
            if branch is None:
                break
            with self.handle_errors():
                self.process_workflow_branch(branch)

    def get_next_workflow_branch(self):
        log.debug('Decider.get_next_workflow_branch()')
        data = self.microserver.get_workflow_to_execute(self.uuid)
        log.debug('Decider.get_next_workflow_branch() => %s', data)
        if data == []:
            return None
        return Branch(
            name=data['name'],
            event=deserialize(data['event']) if data['event'] else None,
            properties=deserialize(data['properties']),
        )

    @contextmanager
    def handle_errors(self):
        try:
            yield
        except ZenatonError:
            etype, value, traceback = sys.exc_info()
            self.microserver.notify_decision_internal_error(
                self.uuid, etype, value, traceback
            )
            raise
        except Exception:
            etype, value, traceback = sys.exc_info()
            self.microserver.notify_decision_user_error(
                self.uuid, etype, value, traceback
            )
            raise

    def process_workflow_branch(self, branch):
        log.debug('Decider.process_workflow_branch(%s)', branch)
        workflow = self.create_workflow(
            name=branch.name,
            properties=branch.properties,
        )
        try:
            output = self.run_handler(workflow, branch.event)
        except ScheduledBoxException:
            self.microserver.notify_decision_complete(
                uuid=self.uuid,
                properties=get_properties(workflow),
            )
            return
        self.microserver.notify_decision_branch_complete(
            uuid=self.uuid,
            properties=get_properties(workflow),
            output=output,
        )

    def create_workflow(self, name, properties):
        log.debug(
            'Decider.create_workflow(name=%s, properties=%s)',
            name,
            properties,
        )
        workflow = Workflow.create(name, properties)
        workflow.__zenaton_executor__ = WorkflowExecutor(
            uuid=self.uuid,
            microserver=self.microserver,
        )
        return workflow

    def run_handler(self, workflow, event):
        log.debug(
            'Decider.run_handler(workflow=%s, event=%s)',
            workflow,
            event,
        )
        if event is not None and hasattr(workflow, 'on_event'):
            return workflow.on_event(event)
        return workflow.handle()



from __future__ import absolute_import, print_function, unicode_literals

import inspect
import logging
from importlib import import_module

from zenaton.errors import ZenatonError


log = logging.getLogger(__name__)


class InstantiationError(ZenatonError):
    pass


class InvalidClassPath(InstantiationError):
    pass


class ImportModuleError(InstantiationError):
    pass


class ClassNotFoundError(InstantiationError):
    pass


class NotAClassError(InstantiationError):
    pass


class InvalidProperty(InstantiationError):
    pass


class ConstructorError(InstantiationError):
    pass


def instantiate_class(class_path, properties):
    """
    Instantiate an object based on module and class names
    """
    try:
        module_name, class_name = class_path.split(':')
    except ValueError:
        raise InvalidClassPath("Invalid class path '{}'".format(class_path))

    try:
        module = import_module(module_name)
    except ImportError:
        raise ImportModuleError("Could not import module '{}'".format(
            module_name
        ))

    try:
        class_ = getattr(module, class_name)
    except AttributeError:
        raise ClassNotFoundError("Could not find class '{}.{}'".format(
            module_name,
            class_name,
        ))

    if not inspect.isclass(class_):
        raise NotAClassError("'{}' is not a class".format(class_name))

    forbidden_keys = [key for key in properties.keys() if key.startswith('_')]
    if forbidden_keys:
        if len(forbidden_keys) == 1:
            msg = "Property name '{}' is not allowed".format(forbidden_keys[0])
        else:
            msg = "Property names {} are not allowed".format(
                ', '.join(["'{}'".format(key) for key in forbidden_keys])
            )
        raise InvalidProperty(msg)

    try:
        obj = class_(**properties)
    except Exception:
        msg = "Could not create object of class '{}'".format(class_name)
        log.exception(msg)
        raise ConstructorError(msg)

    return obj



from __future__ import absolute_import, print_function, unicode_literals

import json

import pytest

from zenaton.serialization import Serializable


class MyClass(Serializable):
    pass


class NotSerializable:
    pass


class TestSerializable:

    def test_init_properties(self):
        obj = MyClass(a=1)
        assert obj.a == 1

    def test_invalid_property_name(self):
        from zenaton.serialization import InvalidPropertyError
        with pytest.raises(InvalidPropertyError):
            obj = MyClass(_a=1)

    def test_create(self):
        obj = Serializable.create('test_serialization:MyClass', {'a': 1})
        assert isinstance(obj, MyClass)
        assert obj.a == 1

    def test_create_not_subclass(self):
        from zenaton.serialization import InvalidClassError

        class Thing(Serializable):
            pass

        with pytest.raises(InvalidClassError) as exc_info:
            Thing.create('test_serialization:MyClass', {'a': 1})
        assert str(exc_info.value) == 'Not a Thing'


class TestSerialize:

    def test_serialize_none(self):
        from zenaton.serialization import serialize
        assert serialize(None) == "null"

    def test_serialize_list(self):
        from zenaton.serialization import serialize
        assert json.loads(serialize([1, 2])) == [1, 2]

    def test_serialize_dict(self):
        from zenaton.serialization import serialize
        assert json.loads(serialize({"a": 1, "b": 2})) == {
            "a": 1,
            "b": 2,
        }

    def test_serialize_serializable_object(self):
        from zenaton.serialization import serialize
        obj = MyClass(a=1, b=2)
        assert json.loads(serialize(obj)) == {
            "name": "test_serialization:MyClass",
            "properties": {"a": 1, "b": 2},
        }

    def test_serialize_non_serializable_object(self):
        from zenaton.serialization import serialize, SerializeError

        with pytest.raises(SerializeError):
            serialize(NotSerializable())


class TestDeserialize:

    def test_deserialize_none(self):
        from zenaton.serialization import deserialize
        assert deserialize(None) is None

    def test_deserialize_null(self):
        from zenaton.serialization import deserialize
        assert deserialize('null') is None

    def test_deserialize_list(self):
        from zenaton.serialization import deserialize
        assert deserialize('[1, 2]') == [1, 2]

    def test_deserialize_dict(self):
        from zenaton.serialization import deserialize
        assert deserialize('{"a": 1}') == {"a": 1}

    def test_deserialize_object(self):
        from zenaton.serialization import deserialize
        obj = deserialize(
            '{"name": "test_serialization:MyClass",'
            ' "properties": {"a": 1, "b": 2}}'
        )
        assert isinstance(obj, MyClass)

    def test_deserialize_error(self):
        from zenaton.serialization import deserialize, DeserializeError
        with pytest.raises(DeserializeError):
            deserialize('invalid')



from __future__ import absolute_import, print_function, unicode_literals

import pytest

from zenaton.testing.compat import MagicMock
from zenaton.version import __version__


@pytest.fixture
def api():
    from zenaton.api import API
    api = API(app_id='APP_ID', token='API_TOKEN', environment='APP_ENV')
    api.session = MagicMock()
    return api


class TestAPIBaseURL:

    def test_api_base_url_default(self, api):
        assert api.base_url == 'https://zenaton.com/api'

    def test_api_base_url_override(self):
        from zenaton.api import API
        api = API(
            app_id='APP_ID',
            token='API_TOKEN',
            environment='APP_ENV',
            base_url='https://localhost:8080/api',
        )
        assert api.base_url == 'https://localhost:8080/api'


class TestAPIWorkflow:

    def test_api_start_workflow(self, api):
        api._post_json = MagicMock()

        res = api.start_workflow('NAME', {'foo': 'bar'}, 'CUSTOM_ID')

        api._post_json.assert_called_once_with(
            'https://zenaton.com/api/instances',
            data={
                'name': 'NAME',
                'data': {'foo': 'bar'},
                'custom_id': 'CUSTOM_ID',
                'programming_language': 'Python',
            }
        )

    def test_api_update_workflow_instance(self, api):
        api._put_json = MagicMock()

        res = api.update_workflow_instance('ID', 'NAME', 'STATE')

        api._put_json.assert_called_once_with(
            'https://zenaton.com/api/instances/ID',
            data={
                'name': 'NAME',
                'programming_language': 'Python',
                'state': 'STATE',
            }
        )

    def test_api_get_workflow_instance(self, api):
        api._get_json = MagicMock()

        res = api.get_workflow_instance('ID', 'NAME')

        api._get_json.assert_called_once_with(
            'https://zenaton.com/api/instances/ID',
            params={
                'name': 'NAME',
                'programming_language': 'Python',
            }
        )


class TestAPISendEvent:

    def test_api_send_event(self, api):
        api._post_json = MagicMock()
        res = api.send_event('ID', 'WORKFLOW', 'EVENT', {'foo': 'bar'})
        api._post_json.assert_called_once_with(
            'https://zenaton.com/api/events',
            data={
                'custom_id': 'ID',
                'event_input': {'foo': 'bar'},
                'event_name': 'EVENT',
                'name': 'WORKFLOW',
                'programming_language': 'Python',
            }
        )


class TestAPIHTTP:

    def test_get_json(self, api):
        api._request_json = MagicMock()
        api._request_json.return_value = {'foo': 'bar'}

        res = api._get_json('URL', params={'a': 1})

        api._request_json.assert_called_once_with(
            'GET',
            'URL',
            params={'a': 1},
        )
        assert res == {'foo': 'bar'}

    def test_post_json(self, api):
        api._request_json = MagicMock()
        api._request_json.return_value = {'foo': 'bar'}

        res = api._post_json('URL', {'a': 1})

        api._request_json.assert_called_once_with(
            'POST',
            'URL',
            data={'a': 1},
            params=None,
        )
        assert res == {'foo': 'bar'}

    def test_put_json(self, api):
        api._request_json = MagicMock()
        api._request_json.return_value = {'foo': 'bar'}

        res = api._put_json('URL', {'a': 1})

        api._request_json.assert_called_once_with(
            'PUT',
            'URL',
            data={'a': 1},
            params=None,
        )
        assert res == {'foo': 'bar'}

    def test_request_json(self, api):
        api._request_json(
            'POST',
            'https://localhost:8080/api',
            data={'data': 1},
            params={'param': 1},
        )
        api.session.request.assert_called_once_with(
            'POST',
            'https://localhost:8080/api',
            params={
                'param': 1,
                'app_env': 'APP_ENV',
                'app_id': 'APP_ID',
                'api_token': 'API_TOKEN',
            },
            headers={
                'Accept': 'application/json',
                'Content-Type': 'application/json',
                'User-Agent': 'zenaton-python/{}'.format(__version__),
            },
            data='{"data": 1}',
        )
        api.session.request.return_value.json.assert_called_once_with()

    def test_request_json_unauthenticated(self, api):
        from zenaton.api import UnauthenticatedError

        api.session.request.return_value.status_code = 401

        with pytest.raises(UnauthenticatedError):
            api._request_json('POST', 'https://localhost:8080/api', data={})

    def test_request_json_error_payload(self, api):
        from zenaton.api import APIError

        api.session.request.return_value.json.return_value = {
            'error': 'some error message',
        }

        with pytest.raises(APIError) as exc:
            api._request_json('POST', 'https://localhost:8080/api', data={})

        assert str(exc.value) == 'some error message'



"""Initialize the microserver.

Usage:
  zenaton_init [options] [--microserver-url=URL]
  zenaton_init -h | --help

Options:
  -h --help     Show this screen.
  -v --verbose  Increase log level.
  -d --debug    Increase log level even more.
"""
from __future__ import absolute_import, print_function, unicode_literals

import logging
import sys

from docopt import docopt

from zenaton.config import Config
from zenaton.microserver import MicroserverAPI
from zenaton.scripts.base import setup_logging


log = logging.getLogger(__name__)


def main(argv=None):
    args = docopt(__doc__, argv=argv)
    setup_logging(verbose=args['--verbose'], debug=args['--debug'])
    try:
        init_microserver(base_url=args['--microserver-url'])
        sys.exit(0)
    except Exception:
        logging.exception("Unexpected error")
        sys.exit(1)


def init_microserver(base_url=None):
    microserver = MicroserverAPI(base_url=base_url)
    log.debug('Sending config to microserver on {}'.format(
        microserver.base_url
    ))

    config = Config().to_dict()
    log.debug(config)

    microserver.send_config(config=config)



from __future__ import absolute_import, print_function, unicode_literals

import logging

from zenaton.errors import ZenatonError
from zenaton.serialization import Serializable
from zenaton.task import Task
from zenaton.tasks.wait import Wait


log = logging.getLogger(__name__)


class NotExecutableError(ZenatonError):
    pass


class Workflow(Serializable):
    """
    Base class for workflows.
    """

    def handle(self):
        """
        You need to implement this method in your derived class.
        """
        raise NotImplementedError  # pragma: nocover

    def execute(self, *executables, **kwargs):
        log.debug('Workflow.execute(args=%s, kwargs=%s)', executables, kwargs)
        for executable in executables:
            if not isinstance(executable, (Task, Workflow, Wait)):
                raise NotExecutableError("'{}' is not a Task or a Workflow")
        async_ = kwargs.pop('async_', False)
        if self._is_running_within_zenaton():
            return self.__zenaton_executor__.execute(self, executables, async_)
        else:
            return self._execute_normal(executables, async_=async_)

    def _is_running_within_zenaton(self):
        return getattr(self, '__zenaton_executor__', None) is not None

    def _execute_normal(self, executables, async_):
        outputs = []
        for executable in executables:
            outputs.append(executable.handle())
        if executables and not async_:
            return outputs if len(outputs) > 1 else outputs[0]

    def execute_async(self, *executables):
        return self.execute(*executables, async_=True)



from __future__ import absolute_import, print_function, unicode_literals

from random import randint
from time import sleep

from zenaton import Task


class SendInvite(Task):

    def __init__(self, name):
        self.name = name

    def handle(self):
        print('Sending invite to: {}'.format(self.name))
        sleep(randint(1, 3))
        print('Invite succesfully sent to: {}'.format(self.name))



from __future__ import absolute_import, print_function, unicode_literals

from zenaton import Workflow

from .address_updated_event import AddressUpdatedEvent
from .prepare_order import PrepareOrder
from .send_order import SendOrder


class OrderWorkflow(Workflow):

    def __init__(self, item, address):
        self.item = item
        self.address = address

    def handle(self):
        self.execute(PrepareOrder(self.item))
        self.execute(SendOrder(self.item, self.address))

    def on_event(self, event):
        if isinstance(event, AddressUpdatedEvent):
            self.address = event.address



from __future__ import absolute_import, print_function, unicode_literals


class ZenatonError(Exception):
    pass


class APIError(ZenatonError):
    pass


class UnauthenticatedError(APIError):
    pass



from __future__ import absolute_import, print_function, unicode_literals

import pytest

from zenaton.errors import ZenatonError
from zenaton.testing.compat import MagicMock
from zenaton.workflow import Workflow


@pytest.fixture
def client():
    from zenaton.client import Client

    client = Client(
        app_id='MY_APP_ID',
        token='MY_TOKEN',
        environment='TEST',
    )
    client.api = MagicMock()
    return client


class WF(Workflow):

    def __init__(self, id=None):
        self._id = id
        self.a = 1

    def get_id(self):
        return self._id


class TestClientStartWorkflow:

    def test_workflow_has_no_custom_id(self, client):
        client.api.start_workflow.return_value = {
            'msg': 'test_client:WF launched',
            'custom_id': 'c8rbrivw188ws4s',
            'instance_id':'c8rbrivw188ws4s-bipmzra7qfsckcc',
        }

        instance = client.start(WF())

        client.api.start_workflow.assert_called_once_with(
            name='test_client:WF',
            data='{"a": 1}',
            custom_id=None,
        )
        assert instance.id == 'c8rbrivw188ws4s'

    def test_workflow_has_a_string_custom_id(self, client):
        client.api.start_workflow.return_value = {
            'msg': 'test_client:WF launched',
            'custom_id': '55gffs7c2ykg804w',
            'instance_id': '55gffs7c2ykg804w-bipmzra7qfsckcc',
        }

        instance = client.start(WF(id='55gffs7c2ykg804w'))

        client.api.start_workflow.assert_called_once_with(
            name='test_client:WF',
            data='{"a": 1}',
            custom_id='55gffs7c2ykg804w',
        )
        assert instance.id == '55gffs7c2ykg804w'

    def test_workflow_has_a_binary_custom_id(self, client):
        client.api.start_workflow.return_value = {
            'msg': 'test_client:WF launched',
            'custom_id': '55gffs7c2ykg804w',
            'instance_id': '55gffs7c2ykg804w-bipmzra7qfsckcc',
        }

        instance = client.start(WF(id=b'custom id'))

        client.api.start_workflow.assert_called_once_with(
            name='test_client:WF',
            data='{"a": 1}',
            custom_id='custom id',
        )
        assert instance.id == '55gffs7c2ykg804w'

    def test_error_workflow_custom_id_is_not_a_string(self, client):
        workflow = WF(id=42)

        with pytest.raises(ZenatonError) as exc:
            instance = client.start(workflow)

        assert str(exc.value) == 'Workflow ID must be a string'

    def test_error_workflow_custom_id_is_too_long(self, client):
        workflow = WF(id=(b'a' * 192))

        with pytest.raises(ZenatonError) as exc:
            instance = client.start(workflow)

        assert str(exc.value) == 'Workflow ID is too long'

    def test_error_not_a_workflow(self, client):
        class NotAWorkflow:
            pass

        with pytest.raises(ZenatonError):
            client.start(NotAWorkflow())



from __future__ import absolute_import, print_function, unicode_literals

from zenaton import Task


class SendBookingConfirmation(Task):

    def __init__(self, request):
        self.request = request

    def handle(self):
        print('Sending notification to customer:')
        print('- Customer ID: {}'.format(self.request['customer_id']))
        print('- Request ID: {}'.format(self.request['id']))
        print('- Car ID: {}'.format(self.request['booking_id']))



from __future__ import absolute_import, print_function, unicode_literals

import pytest

from zenaton.testing.compat import MagicMock


@pytest.fixture
def microserver():
    return MagicMock()


@pytest.fixture
def executor(microserver):
    from zenaton.workflow_executor import WorkflowExecutor

    return WorkflowExecutor(
        uuid='88282738-7A5C-48A5-BEF7-043FD66B7CD1',
        microserver=microserver,
    )


class Task1:
    pass


class Task2:
    pass


class TestBuildBoxes:

    def test_build_boxes_single_sync(self, executor):

        boxes = executor._build_boxes(executables=[Task1], async_=False)

        assert len(boxes) == 1

        assert boxes[0].executable == Task1
        assert boxes[0].position_str == '1'

    def test_build_boxes_multiple_sync(self, executor):

        boxes = executor._build_boxes(executables=[Task1, Task2], async_=False)

        assert len(boxes) == 2

        assert boxes[0].executable == Task1
        assert boxes[0].position_str == '1p0'

        assert boxes[1].executable == Task2
        assert boxes[1].position_str == '1p1'

    def test_build_boxes_single_async(self, executor):

        boxes = executor._build_boxes(executables=[Task1], async_=True)

        assert len(boxes) == 1

        assert boxes[0].executable == Task1
        assert boxes[0].position_str == '1a'

    def test_build_boxes_multiple_async(self, executor):

        boxes = executor._build_boxes(executables=[Task1, Task2], async_=True)

        assert len(boxes) == 2

        assert boxes[0].executable == Task1
        assert boxes[0].position_str == '1a'

        assert boxes[1].executable == Task2
        assert boxes[1].position_str == '2a'



from __future__ import absolute_import, print_function, unicode_literals

from random import randint
from time import sleep

from zenaton import Task


class BookCar(Task):

    def __init__(self, request):
        self.request = request

    def handle(self):
        print('Booking car for request ID: {}'.format(self.request['id']))
        sleep(randint(1, 3))
        self.request['booking_id'] = 'QSFG34'
        # raise Exception('Error Processing Request')
        print('Car booked: {}'.format(self.request['booking_id']))
        return self.request



from __future__ import absolute_import, print_function, unicode_literals

from datetime import datetime
import calendar


def has_method(obj, name):
    return hasattr(obj, name) and callable(getattr(obj, name))


def utc_timestamp():
    return datetime_to_timestamp(datetime.utcnow())


def datetime_to_timestamp(dt):
    return calendar.timegm(dt.utctimetuple())



from __future__ import absolute_import, print_function, unicode_literals

from zenaton.serialization import (
    get_class_path,
    serialize,
)


class Signal:
    KILL = 'kill'
    PAUSE = 'pause'
    RUN = 'run'


class WorkflowProxy:

    def __init__(self, api, id, name):
        self.api = api
        self.id = id
        self.name = name

    def send_event(self, event):
        return self.api.send_event(
            workflow_name=self.name,
            event_name=get_class_path(event.__class__),
            event_input=serialize(event),
            custom_id=self.id,
        )

    def kill(self):
        return self._send_signal(Signal.KILL)

    def pause(self):
        return self._send_signal(Signal.PAUSE)

    def resume(self):
        return self._send_signal(Signal.RUN)

    def _send_signal(self, signal):
        return self.api.update_workflow_instance(self.id, self.name, signal)

    # def get_properties(self):
    #     res = self.api.get_workflow_instance(self.id, self.name)
    #     return deserialize(res['instance'])



from __future__ import absolute_import, print_function, unicode_literals

from collections import OrderedDict

import pytest

from zenaton.testing.compat import patch


class DummyClass:

    def __init__(self, a, b):
        self.a = a
        self.b = b


NOT_A_CLASS = {}


def test_instantiate():
    from zenaton.instantiation import instantiate_class

    obj = instantiate_class('test_instantiation:DummyClass', {
        "a": 1,
        "b": 2,
    })

    assert isinstance(obj, DummyClass)
    assert obj.a == 1
    assert obj.b == 2


@patch('zenaton.instantiation.import_module')
def test_instantiate_import_module_error(mock_import):
    from zenaton.instantiation import instantiate_class, ImportModuleError

    mock_import.side_effect = ImportError

    with pytest.raises(ImportModuleError):
        instantiate_class('non_existing_module:Class', {})


def test_instantiate_class_not_found():
    from zenaton.instantiation import instantiate_class, ClassNotFoundError

    with pytest.raises(ClassNotFoundError):
        instantiate_class('test_instantiation:NonExistingClass', {})


def test_instantiate_not_a_class():
    from zenaton.instantiation import instantiate_class, NotAClassError

    with pytest.raises(NotAClassError):
        instantiate_class('test_instantiation:NOT_A_CLASS', {})


def test_instantiate_invalid_property():
    from zenaton.instantiation import instantiate_class, InvalidProperty

    with pytest.raises(InvalidProperty) as exc_info:
        instantiate_class('test_instantiation:DummyClass', {
            "_a": 1,
            "b": 2,
            "c": 3,
        })

    assert str(exc_info.value) == "Property name '_a' is not allowed"


def test_instantiate_invalid_properties():
    from zenaton.instantiation import instantiate_class, InvalidProperty

    with pytest.raises(InvalidProperty) as exc_info:
        instantiate_class('test_instantiation:DummyClass', OrderedDict([
            ("_a", 1),
            ("b", 2),
            ("_c", 3),
        ]))

    assert str(exc_info.value) == "Property names '_a', '_c' are not allowed"


def test_instantiate_constructor_error():
    from zenaton.instantiation import instantiate_class, ConstructorError

    with pytest.raises(ConstructorError) as exc_info:
        instantiate_class('test_instantiation:DummyClass', {
            "a": 1,
            "b": 2,
            "c": 3,
        })



from __future__ import absolute_import, print_function, unicode_literals

from zenaton.serialization import Serializable


class Task(Serializable):
    """
    Base class for tasks.
    """

    def handle(self):
        """
        You need to implement this method in your derived class.
        """
        raise NotImplementedError  # pragma: nocover



from __future__ import absolute_import, print_function, unicode_literals

from random import randint
from time import sleep

from zenaton import Task


class GetPrice(Task):

    def handle(self):
        # Fake API request to get price from provider B
        print("Contacting provider B to get the price...")
        # raise Exception('Cannot reach provider B')
        sleep(randint(5, 10))
        price = randint(80, 100)
        print('Price from provider B is: {}'.format(price))
        return price


class Order(Task):

    def handle(self):
        print('Ordering "{}" from provider B'.format(self.item['name']))



from __future__ import absolute_import, print_function, unicode_literals

from random import randint
from time import sleep

from zenaton import Task


class PrepareOrder(Task):

    def __init__(self, item):
        self.item = item

    def handle(self):
        print('Preparing order for item {}... '.format(self.item))
        sleep(randint(5, 10))
        print('Order prepared')



from __future__ import absolute_import, print_function, unicode_literals

from zenaton import Workflow

from .provider_a import (
    GetPrice as GetPriceFromProviderA,
    Order as OrderFromProviderA,
)
from .provider_b import (
    GetPrice as GetPriceFromProviderB,
    Order as OrderFromProviderB,
)


class ParallelWorkflow(Workflow):

    def __init__(self, item):
        self.item = item

    def handle(self):
        price_a, price_b = self.execute(
            GetPriceFromProviderA(item=self.item),
            GetPriceFromProviderB(item=self.item),
        )

        if price_a <= price_b:
            self.execute(OrderFromProviderA(item=self.item))
        else:
            self.execute(OrderFromProviderB(item=self.item))



from __future__ import absolute_import, print_function, unicode_literals


class Position:

    SEP_ASYNC = 'a'
    SEP_PARALLEL = 'p'

    def __init__(self):
        self.main = 0
        self.position_str = '0'
        self.counter = 0

    def __str__(self):
        return self.position_str

    def __repr__(self):
        return '<Position({})>'.format(repr(self.position_str))

    def next(self):
        self.main += 1
        self.position_str = str(self.main)

    def next_parallel(self):
        if self.is_parallel():
            self.counter += 1
        else:
            self.main += 1
            self.counter = 0
        self.position_str = '{}{}{}'.format(
            self.main,
            self.SEP_PARALLEL,
            self.counter,
        )

    def is_parallel(self):
        return self.position_str.find(self.SEP_PARALLEL) >= 0

    def next_async(self):
        self.main += 1
        self.position_str = '{}{}'.format(
            self.main,
            self.SEP_ASYNC,
        )



from __future__ import absolute_import, print_function, unicode_literals

from distutils.spawn import find_executable
import os

import six


PROGRAMMING_LANGUAGE = 'Python'
WORKER_SCRIPT = 'zenaton_slave'


class Config:

    @property
    def app_id(self):
        return os.environ['ZENATON_APP_ID']

    @property
    def app_env(self):
        return os.environ['ZENATON_APP_ENV']

    @property
    def api_token(self):
        return os.environ['ZENATON_API_TOKEN']

    @property
    def concurrent_max(self):
        return os.getenv('ZENATON_CONCURRENT_MAX', 100)

    @property
    def _handle_only(self):
        return os.getenv('ZENATON_HANDLE_ONLY', [])

    @property
    def workflows_name_only(self):
        return []  # FIXME

    @property
    def tasks_name_only(self):
        return []  # FIXME

    @property
    def workflows_name_except(self):
        return []  # FIXME

    @property
    def tasks_name_except(self):
        return []  # FIXME

    @property
    def worker_script(self):
        return find_executable(WORKER_SCRIPT)

    def to_dict(self):
        return {
            'app_id': self.app_id,
            'app_env': self.app_env,
            'api_token': self.api_token,
            'concurrent_max': six.text_type(self.concurrent_max),
            'workflows_name_only': self.workflows_name_only,
            'tasks_name_only': self.tasks_name_only,
            'workflows_name_except': self.workflows_name_except,
            'tasks_name_except': self.tasks_name_except,
            'programming_language': PROGRAMMING_LANGUAGE,
            'worker_script': self.worker_script,
            'autoload_path': None,
        }



from __future__ import absolute_import, print_function, unicode_literals

from random import randint
from time import sleep

from zenaton import Task


class SendWelcomeEmail1(Task):

    def __init__(self, email):
        self.email = email

    def handle(self):
        print('Sending welcome email #1 to: {}'.format(self.email))
        sleep(randint(1, 3))
        print('Email #1 sent')



from __future__ import absolute_import, print_function, unicode_literals

from zenaton import Event


class AddressUpdatedEvent(Event):

    def __init__(self, address):
        self.address = address



from __future__ import absolute_import, print_function, unicode_literals

from distutils.version import LooseVersion


def test_version_number_is_valid():
    from zenaton.version import __version__
    assert __version__
    assert LooseVersion(__version__)



from __future__ import absolute_import, print_function, unicode_literals

from client import client
from sequential.car_booking_workflow import CarBookingWorkflow


def main():
    request = {
        'id': 12345,
        'customer_id': '2DER45G',
        'transport': 'car',
    }
    workflow = CarBookingWorkflow(request)

    # Direct synchronous execution
    # workflow.handle()

    # Execution through Zenaton
    instance = client.start(workflow)
    print('Launched workflow instance {}'.format(instance.id))


if __name__ == '__main__':
    main()



from __future__ import absolute_import, print_function, unicode_literals

import pytest

from zenaton.event import Event
from zenaton.workflow import Workflow

from zenaton.testing.compat import MagicMock, call, patch


class MyWorkflow(Workflow):
    def __init__(self, a):
        self.a = a


class MyEvent(Event):
    pass


@pytest.fixture
def decider():
    from zenaton.decider import Decider
    return Decider(
        uuid='F26C7548-F6AB-4C2C-A5EB-745D8BD38D61',
        microserver=MagicMock(),
    )


class TestLaunch:

    def test_launch(self, decider):
        mock_branch1 = MagicMock()
        mock_branch2 = MagicMock()

        decider.get_next_workflow_branch = MagicMock()
        decider.get_next_workflow_branch.side_effect = [
            mock_branch1,
            mock_branch2,
            None,
        ]

        decider.process_workflow_branch = MagicMock()

        decider.launch()

        decider.process_workflow_branch.assert_has_calls([
            call(mock_branch1),
            call(mock_branch2),
        ])

    def test_launch_nothing_to_run(self, decider):
        decider.get_next_workflow_branch = MagicMock()
        decider.get_next_workflow_branch.return_value = None

        decider.process_workflow_branch = MagicMock()

        decider.launch()

        assert not decider.process_workflow_branch.called


def test_get_next_workflow_branch(decider):
    from zenaton.decider import Branch

    decider.microserver.get_workflow_to_execute.return_value = {
        'name': 'MODULE:CLASS',
        'properties': '{"a": 1}',
        'event': '{"name":"test_decider:MyEvent","properties":{}}',
    }
    branch = decider.get_next_workflow_branch()

    decider.microserver.get_workflow_to_execute.assert_called_once_with(
        'F26C7548-F6AB-4C2C-A5EB-745D8BD38D61'
    )
    assert isinstance(branch, Branch)
    assert branch.name == 'MODULE:CLASS'
    assert branch.properties == {"a": 1}
    assert isinstance(branch.event, MyEvent)


class TestHandleErrors:

    def test_handle_errors_ok(self, decider):

        with decider.handle_errors():
            pass

        m_notify_user = decider.microserver.notify_decision_user_error
        assert not m_notify_user.called

        m_notify_internal = decider.microserver.notify_decision_internal_error
        assert not m_notify_internal.called

    @patch('sys.exc_info')
    def test_handle_errors_internal_error(self, mock_exc_info, decider):
        from zenaton.errors import ZenatonError

        mock_exc_info.return_value = ('ETYPE', 'VALUE', 'TB')

        with pytest.raises(ZenatonError):
            with decider.handle_errors():
                raise ZenatonError

        mock_notify = decider.microserver.notify_decision_internal_error
        mock_notify.assert_called_once_with(
            'F26C7548-F6AB-4C2C-A5EB-745D8BD38D61',
            'ETYPE',
            'VALUE',
            'TB',
        )

    @patch('sys.exc_info')
    def test_handle_errors_user_error(self, mock_exc_info, decider):

        mock_exc_info.return_value = ('ETYPE', 'VALUE', 'TB')

        with pytest.raises(RuntimeError):
            with decider.handle_errors():
                raise RuntimeError

        mock_notify = decider.microserver.notify_decision_user_error
        mock_notify.assert_called_once_with(
            'F26C7548-F6AB-4C2C-A5EB-745D8BD38D61',
            'ETYPE',
            'VALUE',
            'TB',
        )


class TestProcessWorkflowBranch:

    def test_process_workflow_branch(self, decider):
        from zenaton.decider import Branch

        m_workflow = MyWorkflow(a=1)

        decider.create_workflow = MagicMock()
        decider.create_workflow.return_value = m_workflow

        decider.run_handler = MagicMock()
        decider.run_handler.return_value = 'OUTPUT'

        m_event = MyEvent()

        decider.process_workflow_branch(
            branch=Branch(
                name='MODULE:CLASS',
                properties={"a": 1},
                event=m_event,
            )
        )

        decider.create_workflow.assert_called_once_with(
            name='MODULE:CLASS',
            properties={"a": 1},
        )

        decider.run_handler.assert_called_once_with(m_workflow, m_event)

        mock_notify = decider.microserver.notify_decision_branch_complete
        mock_notify.assert_called_once_with(
            uuid='F26C7548-F6AB-4C2C-A5EB-745D8BD38D61',
            properties={
                'a': 1,
            },
            output='OUTPUT',
        )

        mock_notify_complete = decider.microserver.notify_decision_complete
        assert not mock_notify_complete.called

    def test_process_workflow_branch_box_exception(self, decider):
        from zenaton.decider import Branch, ScheduledBoxException

        decider.create_workflow = MagicMock()
        decider.create_workflow.return_value = MyWorkflow(a=1)

        decider.run_handler = MagicMock()
        decider.run_handler.side_effect = ScheduledBoxException

        m_event = MyEvent()

        decider.process_workflow_branch(
            branch=Branch(
                name='MODULE:CLASS',
                properties={"a": 1},
                event=m_event,
            )
        )

        mock_notify = decider.microserver.notify_decision_branch_complete
        assert not mock_notify.called

        mock_notify_complete = decider.microserver.notify_decision_complete
        mock_notify_complete.assert_called_once_with(
            uuid='F26C7548-F6AB-4C2C-A5EB-745D8BD38D61',
            properties={
                'a': 1,
            }
        )



import pytest


@pytest.fixture
def position():
    from zenaton.position import Position

    return Position()


def test_default(position):
    assert str(position) == '0'


def test_next(position):
    position.next()
    assert str(position) == '1'

    position.next()
    assert str(position) == '2'


def test_next_parallel(position):
    position.next_parallel()
    assert str(position) == '1p0'

    position.next_parallel()
    assert str(position) == '1p1'



"""
End-to-end test with fake remote and microserver APIs
"""
from __future__ import absolute_import, print_function, unicode_literals

import json

import responses

from zenaton.testing.compat import patch


@patch('sys.exit')
def test_zenaton_slave_run_task(mock_exit, capsys):
    from zenaton.scripts.slave import main

    with responses.RequestsMock() as rsps:

        # Get a job to do => run a task
        rsps.add(
            responses.GET,
            'http://localhost:4001/jobs/AD489DB1',
            status=200,
            json={
                'action': 'TaskScheduled',
                'uuid': '547BDD74-CE71-4762-96F6-94F5DC3CFD68',
                'name': 'zenaton.testing.examples.readme:Echo',
                'input': '{"value": "Hello world"}',
                'hash': '8749975af5306c045c5c0af268a29482',
            }
        )

        # Notify task completion => OK
        rsps.add(
            responses.POST,
            'http://localhost:4001/works/547BDD74-CE71-4762-96F6-94F5DC3CFD68',
            status=200,
            body='{}',
            content_type='application/json',
         )

        main(['AD489DB1', '8CF89842'])

        assert len(rsps.calls) == 2

        assert rsps.calls[0].request.url == 'http://localhost:4001/jobs/AD489DB1?slave_id=8CF89842'

        assert rsps.calls[1].request.url == 'http://localhost:4001/works/547BDD74-CE71-4762-96F6-94F5DC3CFD68'
        assert json.loads(rsps.calls[1].request.body) == {
            "action": "terminate",
            "status": "completed",
            "output": "null",
            "duration": 0,
            "hash": '8749975af5306c045c5c0af268a29482',
        }

    # Check that the task printed its argument
    out, err = capsys.readouterr()
    assert out == 'Hello world\n'
    assert err == ''

    mock_exit.assert_called_once_with(0)



from __future__ import absolute_import, print_function, unicode_literals

from time import sleep, time


class Wait:

    def __init__(self, seconds=0, event=None):
        self.seconds = seconds
        self.timeout_timestamp = int(time() + self.seconds)
        self.event = event

    def handle(self):
        sleep(self.seconds)

    def get_event(self):
        return self.event

    def get_timeout_timestamp(self):
        return self.timeout_timestamp
